import {AbstractPayment, SlpPayments, SlpPaymentsApiRes} from "./SlpPayments";
import {WebHelpers} from "./WebHelpers";
import {AbstractModule} from "./AbstractModule";

export interface BadgerWalletPayment extends AbstractPayment {
    buttonId: string; // unique ID, tx ID in our case
    txid: string;
    amount: number;
    tokenAmount?: number;
    currency: string;
    buttonData: string; // JSON string "{}" or base64 data in our app
    buttonDataObj: any; // unserialized JS object from buttonData
}

export class BadgerWallet extends AbstractModule {
    protected static readonly INSTALL_URL = "https://badger.bitcoin.com/";

    protected globalCallbacks: boolean; // expose callback functions to window object because BadgerButton currently doesn't support functions on objects

    constructor(slpPayments: SlpPayments, webHelpers: WebHelpers, globalCallbacks = false) {
        super(slpPayments, webHelpers);
        this.globalCallbacks = globalCallbacks;

        if (this.globalCallbacks === true) {
            setTimeout(() => { // must be done after constructor or else functions are undefined
                this.slpPayments.window['onSlpBadgerPayment'] = this.onBadgerPayment;
                this.slpPayments.window['onSlpBadgerClientPayment'] = this.sendPaymentReceived;
            }, 0);
        }

        this.slpPayments.$(this.slpPayments.window.document).ready(($) => {
            this.addEventListeners();
        });
    }

    public isInstalled() {
        return (this.slpPayments.window as any).web4bch !== undefined;
    }

    public isLoggedIn() {
        const wnd: any = this.slpPayments.window;
        return typeof wnd.web4bch.bch.defaultAccount === "string" && wnd.web4bch.bch.defaultAccount !== "";
    }

    // ################################################################
    // ###################### PRIVATE FUNCTIONS #######################

    protected onBadgerPayment(payment: BadgerWalletPayment) {
        if (typeof payment.buttonData === "string" && payment.buttonData.length !== 0 && payment.buttonData[0] !== "{")
            payment.buttonDataObj = JSON.parse(this.slpPayments.getWebHelpers().fromBase64(payment.buttonData)) || {}; // this.webHelpers is undefined
        if (payment.domID === undefined)
            payment.domID = "slp-btn-wrap-" + payment.buttonId;
        //console.log("Badger Payment", payment);

        //this.sendPaymentReceived(payment);
        this.slpPayments.window['onSlpBadgerClientPayment'].call(this.slpPayments.badger, payment, (res: SlpPaymentsApiRes) => {
            if (payment.buttonDataObj && payment.buttonDataObj.woocommerce === true)
                this.slpPayments.woocommerce.onPayment(payment);
        });
        // TODO increment tips received and donation goal progress with JS
        if (typeof this.slpPayments.window['onSlpPayment'] === "function")
            this.slpPayments.window['onSlpPayment'].call(this.slpPayments.window, {badger: payment});
    }

    protected addEventListeners() {
        this.slpPayments.$(".slp-badger-button").click((event) => {
            if (this.isInstalled() === false)
                this.slpPayments.window.open(BadgerWallet.INSTALL_URL, "", "");
            else if (this.isLoggedIn() === false) {
                event.preventDefault();
                this.slpPayments.window.alert(this.slpPayments.getConfig().badgerLocked);
            }
            else
                this.sendPayment(event.target);
        });
    }

    protected sendPayment(button: Element) {
        const wnd: any = this.slpPayments.window;
        let web4bch = wnd.web4bch;
        web4bch = new wnd.Web4Bch(web4bch.currentProvider);

        const btn = this.slpPayments.$(button);
        const tokenAmount = btn.attr("data-tokens");
        let txParams: any = {
            to: btn.attr("data-to"),
            from: web4bch.bch.defaultAccount,
            //value: btn.attr("data-satoshis") // TODO wait for ability to send tokens + BCH
            value: tokenAmount ? tokenAmount : btn.attr("data-satoshis")
        }
        if (tokenAmount) {
            txParams.sendTokenData = {
                tokenId: btn.attr("data-token-id"),
                tokenProtocol: 'slp'
            }
        }

        web4bch.bch.sendTransaction(txParams, (err, res) => {
            if (err) {
                this.slpPayments.window.console.log("Error sending payment", err);
                return;
            }
            // call the corresponding inline-javascript for this button
            let callback = btn.attr("data-success-callback");
            if (callback && typeof this.slpPayments.window[callback] === "function")
                this.slpPayments.window[callback](res); // res is just the txid as string
        });
    }

    protected sendPaymentReceived(payment: BadgerWalletPayment, callback?: (res: SlpPaymentsApiRes) => void) {
        let params = {
            txid: payment.buttonId,
            am: payment.amount,
            amT: payment.tokenAmount,
            keep: this.slpPayments.getConfig().keepTransaction === true
        }
        this.webHelpers.getApi("/wp-json/slp-payments/v1/badger-client", params, (data) => {
            callback && callback(data);
        });
    }
}
