import {AbstractPayment, SlpPayments} from "./SlpPayments";
import {WebHelpers} from "./WebHelpers";
import {Order} from "./structs/Order";
import {AbstractModule} from "./AbstractModule";
import {SlpDb} from "./api/SlpDb";

/**
 * Class to interact with the Woocommerce store of a WordPress installation.
 */
export class Woocommerce extends AbstractModule {
    protected static readonly MAX_TRY_ADD_PAY_FORM_LISTENER_COUNT = 5;
    protected static readonly ADD_CHECKOUT_FORM_LISTENERS_DELAY_MS = 500;

    //protected updateFormTimerID: number = 0;
    protected slpDb = new SlpDb();
    protected receivedAmount: number = 0.0; // received amount for current order checked via SLPDB. used to trigger server-side re-check
    protected fullyPaid: boolean = false;
    protected checkServerPaymentTimerID: number = 0;
    protected slpEventStream: EventSource = null;

    constructor(slpPayments: SlpPayments, webHelpers: WebHelpers) {
        super(slpPayments, webHelpers);
        this.slpPayments.$(this.slpPayments.window.document).ready(($) => {
            setTimeout(() => {
                const config = this.slpPayments.getConfig();
                // check payment stastus sooner on 1st page load to update remaining amount
                if (this.slpPayments.$("#slp-payment-options").length !== 0 && config.checkPaymentIntervalSec > 0) {
                    if (this.webHelpers.hasEventSourceSupport() === true && typeof this.slpPayments.window.btoa === "function")
                        setTimeout(this.checkPaymentStatusViaStream.bind(this, true), /*config.checkPaymentIntervalSec*/1 * 1000);
                    else
                        setTimeout(this.checkPaymentStatus.bind(this, true), /*config.checkPaymentIntervalSec*/1 * 1000);
                }

                if (this.slpPayments.$("#slp-token-infobox").length === 0)
                    return;
                //this.scheduleUpdateRemainingAmount();
                this.tryAddPayFormListener(); // it might be the first payment method (already selected)
            }, 1000);
            this.slpPayments.$('form.checkout input[name^="payment_method"]').on("change", (event) => {
                this.addPayFormListener(); // the form before the order is placed
            });

            if (this.slpPayments.$("#slp-qrcode-form").length !== 0)
                this.addPaymentFormEvents(this.slpPayments.$("#slp-qrcode-form").eq(0)); // the form after the order has been placed
        });

        this.slpPayments.$(this.slpPayments.window).load(($) => {
            this.addButtonLinkClass();
            if (this.slpPayments.$(".slp-preloader").length > 0) {
                this.slpPayments.$(".slp-preloader").fadeOut("slow");
            }
        });
    }

    public onPayment(payment: AbstractPayment) {
        this.sendPaymentForValidation(payment);
    }

    // ################################################################
    // ###################### PRIVATE FUNCTIONS #######################

    protected sendPaymentForValidation(payment: AbstractPayment) {
        const firstButton = this.slpPayments.$(".slp-btn-wrap-top"); // there can only be 1 Woocommerce payment button per payment method
        let params: any = {
            txid: payment.txid,
            dbtxid: this.slpPayments.getPluginPaymentID(firstButton)
            //address: payment.address
            //am: payment.amount
        }
        //if (payment.tokenAmount > 0.0)
            //params.amT = payment.tokenAmount;
        this.webHelpers.getApi("/wp-json/slp-payments/v1/validate", params, (data) => {
            if (data.data.length === 1) {
                this.slpPayments.window.document.location.assign(data.data[0].url);
            }
            // don't show an error message otherwise, because we likely just wait for more confirmations
        });
    }

    protected tryAddPayFormListener(tryCount: number = 0) {
        tryCount++;
        if (this.slpPayments.$("input[name='slp-payment-method']").length === 0) { // form not ready yet (WC loads it async)
            if (tryCount >= Woocommerce.MAX_TRY_ADD_PAY_FORM_LISTENER_COUNT)
                return;
            setTimeout(() => {
                this.tryAddPayFormListener(tryCount);
            }, Woocommerce.ADD_CHECKOUT_FORM_LISTENERS_DELAY_MS);
            return;
        }
        setTimeout(() => { // failed on divi-child theme when adding immediately
            this.slpPayments.$("input[name='slp-payment-method']").eq(0).prop("checked", true);
            this.addPayFormListener();
        }, Woocommerce.ADD_CHECKOUT_FORM_LISTENERS_DELAY_MS);
    }

    protected addPayFormListener() {
        // re-add events because this function can be called multiple times if the user switches payment methods
        /*
        this.slpPayments.$("#tokenAmount").unbind("change").change((event) => {
            this.scheduleUpdateRemainingAmount();
        });
        this.slpPayments.$("#tokenAmount").unbind("keyup").keyup((event) => {
            this.scheduleUpdateRemainingAmount();
        });
        this.slpPayments.$("#tokenAmount").unbind("focus").focus((event) => {
            //this.slpPayments.$("#tokenAmount").focus(); // endless recursion
            this.slpPayments.$("#tokenAmount").select();
        });
        this.slpPayments.$("#payFullToken").unbind("click").click((event) => {
            this.payFullAmountInTokens();
        });
         */
        this.slpPayments.$("input[name='slp-payment-method']").unbind("change").on("change", (event) => {
            this.onSelectedTokenChange();
        });
        this.onSelectedTokenChange();
    }

    protected onSelectedTokenChange() {
        let selectedTokenSymbol = this.getSelectedTokenOnPlaceOrder();
        let selectedToken = this.slpPayments.getConfig().enabledTokens[selectedTokenSymbol]; // is undefined if the user pays in BCH
        if (!selectedToken)
            this.slpPayments.$("#tokenRate").fadeOut("slow");
        else {
            this.slpPayments.$("#tokenRate").fadeIn("slow");
            this.slpPayments.$("#slp-token-rate").text(selectedToken.rate);
        }

        if (selectedToken)
            this.slpPayments.$(".slp-token-symbol").text(selectedToken.name + " (" + selectedToken.symbol + ")");
        if (!selectedToken || this.slpPayments.getWebHelpers().isValidUrl(selectedToken.documentUri) === false) { // hide it if a token without website is selected
            this.slpPayments.$("#slp-token-infobox").fadeOut("slow");
        }
        else {
            this.slpPayments.$("#slp-token-infobox").fadeIn("slow");
            this.slpPayments.$("#slp-token-website").attr("href", selectedToken.documentUri);

            const imageSelector = ".slp-token-img";
            if (!selectedToken.icon || this.slpPayments.getWebHelpers().isValidUrl(selectedToken.icon) === false)
                this.slpPayments.$(imageSelector).fadeOut("slow");
            else {
                this.slpPayments.$(imageSelector).fadeIn("slow");
                this.slpPayments.$(imageSelector).attr("src", selectedToken.icon)
                    .attr("alt", selectedToken.symbol)
                    .attr("title", selectedToken.symbol);
            }
        }
    }

    protected getSelectedTokenOnPlaceOrder(): string {
        let selectedToken = this.slpPayments.$("input[name='slp-payment-method']:checked");
        let symbol = (selectedToken.val() || "").replace(/^slp-/, "");
        return symbol;
    }

    /*
    protected scheduleUpdateRemainingAmount() {
        clearTimeout(this.updateFormTimerID); // delay it to avoid too frequent DOM changes
        this.updateFormTimerID = setTimeout(() => {
            if (this.slpPayments.$("#remainingBchAmount").length === 0)
                return; // removed. currently we only pay full amount in BCH or SLP
            let currentInput = this.slpPayments.$("#tokenAmount").val().replace(/[^0-9\.]+/, "").trim();
            if (currentInput.length === 0)
                this.slpPayments.$("#tokenAmount").val("0.0");
            else
                this.slpPayments.$("#tokenAmount").val(currentInput);
            this.slpPayments.$("#remainingBchAmount").text(this.getRemainingBchAmount().toFixed(this.slpPayments.getConfig().display_digits_bch));
        }, 100);
    }

    protected getRemainingBchAmount(): number {
        let currency = this.slpPayments.getConfig().display_currency;
        let tokenAmount = parseFloat(this.slpPayments.$("#tokenAmount").val()) || 0.0;
        const total = this.slpPayments.getConfig().totalCartValue;
        const remainingAmountBase = Math.max(0.0, total - tokenAmount*this.slpPayments.getConfig().tokenRate[currency]);
        const remainingAmountBch = remainingAmountBase/this.slpPayments.getConfig().rate[currency];
        return remainingAmountBch;
    }

    protected getTokenAmount(fiatAmount: number): number {
        let currency = this.slpPayments.getConfig().display_currency;
        const tokenAmount = fiatAmount / this.slpPayments.getConfig().tokenRate[currency];
        return tokenAmount;
    }
     */

    protected addPaymentFormEvents(paymentFormEl: JQuery) {
        //const txid = paymentFormEl.find("#slp-txid-db").val();

        this.slpPayments.$(".slp-copy-field").click((event) => {
            event.preventDefault();
            const inputEl = this.slpPayments.$(event.target).parent().parent().find("input[type='text']");
            //let text = inputEl.val();
            inputEl.select();
            try {
                this.slpPayments.window.document.execCommand("copy");
            }
            catch (err) { // try-catch shouldn't be needed except for some really old browsers
            }
            inputEl.select(); // ensure it's selected again for iOS devices to copy it manually. not working, remove?
        });
        this.slpPayments.$("#slp-qrcode-form input[type='text']").click((event) => {
            this.slpPayments.$(event.target).select();
        });
        if (this.webHelpers.isAppleIOS() === true) // copy to clipboard button doesn't work there
            this.slpPayments.$("#slp-qrcode-form .slp-copy-field").addClass("hidden");
    }

    /*
    protected payFullAmountInTokens() {
        const total = this.slpPayments.getConfig().totalCartValue;
        const totalToken = this.getTokenAmount(total);
        this.slpPayments.$("#tokenAmount").val(totalToken.toFixed(this.slpPayments.getConfig().display_digits_bch));
        this.scheduleUpdateRemainingAmount();
    }
     */

    protected addButtonLinkClass() {
        const buttons = this.slpPayments.$("#slp-pay-app");
        if (buttons.length === 0)
            return;
        // try to find a native WP theme button style first
        if (this.webHelpers.isExistingCssSelector("a.btn") === true)
            buttons.addClass("btn");
        else if (this.webHelpers.isExistingCssSelector("a.button") === true)
            buttons.addClass("button");
        else
            buttons.addClass("slp-button-link");
    }

    /**
     * Check the payment status
     * @param repeat Whether to repeat the check periodically until the amount is fully paid.
     * @param forceReCheck Force an immediate re-check from our server to ask the Blockchain backend
     */
    protected checkPaymentStatus(repeat = true, forceReCheck = false) {
        const config = this.slpPayments.getConfig();
        if (config.nonce === undefined || config.orderID <= 0)
            return;
        let params: any = {
            n: config.nonce,
            oid: config.orderID
        }
        if (forceReCheck === true)
            params.f = "1";
        clearTimeout(this.checkServerPaymentTimerID);
        this.webHelpers.getApi("/wp-json/slp-payments/v1/order-status", params, (data) => {
            // TODO continue querying if we get sth other than JSON?
            if (repeat === true && config.checkPaymentIntervalSec > 0 && this.fullyPaid === false) { // TODO abort checking after x minutes?
                clearTimeout(this.checkServerPaymentTimerID); // ensure we don't have multiple timers running
                this.checkServerPaymentTimerID = setTimeout(this.checkPaymentStatus.bind(this, true, forceReCheck), config.checkPaymentIntervalSec * 1000); // always repeat the check even if there was an error
            }
            if (data.error === true) {
                this.slpPayments.window.console.error("Error checking SLP payment status: %s", data.errorMsg);
                return;
            }
            if (data.data && data.data.length >= 1) {
                const order: Order = Object.assign(new Order(), data.data[0]);
                if (order.status === "paid") {
                    this.showPaymentReceived();
                    return;
                }
                else if (order.bchAmountReceived > 0.0 || order.tokenAmountReceived > 0.0) {
                    this.showPartialPayment(order);
                    return;
                }
            }
        });
    }

    protected checkPaymentStatusViaStream(repeat = true) {
        const tokenID = this.slpPayments.getConfig().tokenID;
        const address = this.slpPayments.getConfig().address;
        if (address.indexOf("simpleledger") !== 0)
            return this.checkPaymentStatus(repeat); // SLPDB can only be queried for SLP addresses

        const eventUrl = this.slpPayments.getConfig().slpSocketAddress + this.webHelpers.toBase64(this.slpDb.getAddressBalanceQuery(address, tokenID));
        this.slpEventStream = this.webHelpers.listenForServerEvent(eventUrl, (data) => {
            if (repeat === false || this.fullyPaid === true) {
                this.slpEventStream.close();
                this.slpEventStream = null;
            }
            const previousAmount = this.receivedAmount;
            this.receivedAmount += this.slpDb.getTotalOutputAmount(data, address, tokenID);
            //if (this.receivedAmount >= this.slpPayments.getConfig().totalCartValue) // just check every time amount changes, no need to convert USD rates
            if (this.receivedAmount !== previousAmount) {
                // SLLPDB seems a bit slower than SLPSocket, check twice
                clearTimeout(this.checkServerPaymentTimerID);
                this.checkServerPaymentTimerID = setTimeout(this.checkPaymentStatus.bind(this, false, true), 500);
                this.checkServerPaymentTimerID = setTimeout(() => {
                    if (this.fullyPaid === false)
                        this.checkPaymentStatus(false, true);
                }, 1500);
            }
        });
    }

    protected showPartialPayment(order: Order) {
        const numDecimals = order.isTokenPayment === true ? this.slpPayments.getConfig().availableTokens[order.tokenSymbol].decimals : this.slpPayments.getConfig().display_digits_bch;
        const remainingAmount = order.calculateRemaningAmount().toFixed(numDecimals);
        if (this.slpPayments.$("#slp-token-amount-remaining").text() === remainingAmount)
            return;
        this.slpPayments.$("#slp-token-amount-remaining").text(remainingAmount);
        this.slpPayments.$("#slp-remaining-txt").fadeIn("slow");

        // update payment amount in URI and QR code
        this.slpPayments.$("#slp-qr-code-image").attr("src", order.qrcode);
        this.slpPayments.$("#slp-address").val(order.uri);
        this.slpPayments.$("#slp-pay-app").attr("href", order.uri);
        if (order.isTokenPayment === true) // update the amount, the address stays the same
            this.slpPayments.$(".slp-badger-button").attr("data-tokens", order.calculateRemaningAmount());
        else
            this.slpPayments.$(".slp-badger-button").attr("data-satoshis", order.calculateRemaningAmount());
    }

    protected showPaymentReceived() {
        if (this.fullyPaid === true)
            return; // already updated
        this.fullyPaid = true;
        if (this.slpEventStream !== null) { // ensure stream is closed
            this.slpEventStream.close();
            this.slpEventStream = null;
        }
        clearTimeout(this.checkServerPaymentTimerID);
        let successImage = this.slpPayments.window.document.createElement("img");
        successImage.src = this.slpPayments.getConfig().successPayImage;
        successImage.alt = this.slpPayments.getConfig().paidTxt;
        successImage.id = "slp-success-image-animated";
        this.addSuccessImageProperties(successImage);
        setTimeout(() => {
            //let successImageFinal: HTMLImageElement = Object.assign(this.slpPayments.window.document.createElement("img"), successImage); // doesn't copy attributes of native elements
            let successImageFinal = this.slpPayments.window.document.createElement("img");
            successImageFinal.src = this.slpPayments.getConfig().successPayImageFinal; // TODO preload and hide this image URL to prevent flickering on image change
            //successImageFinal.alt = this.slpPayments.getConfig().paidTxt;
            successImageFinal.id = "slp-success-image";
            this.addSuccessImageProperties(successImageFinal);
            this.slpPayments.$("#slp-success-image-animated").replaceWith(successImageFinal);
        }, 1300);

        this.slpPayments.$("#slp-payment-status").text(this.slpPayments.getConfig().paidTxt);
        this.slpPayments.$("#slp-payment-pending, #slp-payment-options, #slp-pay-instructions").fadeOut("slow");
        this.slpPayments.$("#slp-pay-qr-code").replaceWith(successImage);
    }

    protected addSuccessImageProperties(image: HTMLImageElement) {
        image.style.maxWidth = "300px";
        image.style.maxHeight = "300px";
        image.style.width = "50%";
        image.style.display = "block";
        image.className = "slp-center";
    }
}
