<?php
/*
 * This file provides a REST API implementation for automatically updating token rates from CryptoPhyl exchange for this plugin.
 * See:
 * - http://cryptophyl.com/
 * You must upload this file on your server and enter the URL in plugin settings under 'SLP token rate API'.
 * The response JSON to a GET request from WordPress must have the following format:
 * {
	version: 1,
	data: {
		SPICE: {
			price_usd: 0.001
		},
		WPT: {
			price_usd: 1.23
		}
	}
	}
 */

header('Content-Type: application/json; charset=UTF-8');
setCacheHeaders(10); // WP cron calls this file hourly, so we could even set cache headers to 60min
echo json_encode(getSlpTokenRates());


function getSlpTokenRates() {
	$rateUpdater = new FiatRateUpdater();
	$spiceRateGBP = $rateUpdater->getTokenRate('https://api.cryptophyl.com/products/SPICE-BCH/ticker', 'GBP');
	return array(
			'version' => 1,
			'data' => array(
					// in case the returned rate is not positive the SLP plugin will keep the last rate
					'SPICE' => array('price_usd' => $spiceRateGBP, 'price_gbp' => $spiceRateGBP), // _usd included as legacy fallback
			)
	);
}

function setCacheHeaders(int $expires_min) {
	$expires = $expires_min*60;
	header('Pragma: public');
	header('Cache-Control: public, max-age=' . $expires);
	header('Expires: ' . gmdate('D, d M Y H:i:s', time()+$expires) . ' GMT');
}


abstract class SlpRateUpdater {
	protected $rateCache = array(); // assoc array (url with ticker, rate)
	
	public function getBitcoinCashRate(string $baseCurrency): float {
		$this->checkSupportedCurrency($baseCurrency);
		$url = 'https://index-api.bitcoin.com/api/v0/cash/price/' . strtolower($baseCurrency);
		if (isset($this->rateCache[$url]))
			return $this->rateCache[$url];
		$res = $this->httpGet($url);
		if (empty($res)) {
			return -1.0;
		}
		$json = json_decode($res);
		if ($json === null || empty($json->price)) {
			return -1.0;
		}
		if ($json->price > 0.0) {
			$rateUsd = $json->price / 100.0; // receiced in USD cents
			$this->rateCache[$url] = $rateUsd;
			return $rateUsd;
		}
		return -1.0;
	}
	
	protected function httpGet(string $url, array $options = array()) {
		$ctx = stream_context_create(array('http' => 
				array('timeout' => isset($options['timeout']) ? $options['timeout'] : 10,
						'user_agent' => isset($options['userAgent']) ? $options['userAgent'] : "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:61.0) Gecko/20100101 Firefox/61.0",
						'max_redirects' => isset($options['maxRedirects']) ? $options['maxRedirects'] : 5
				))
		);
		$contents = file_get_contents($url, 0, $ctx);
		return $contents;
	}
	
	protected function checkSupportedCurrency(string $baseCurrency): void {
		$supportedCurrencies = array('USD', 'GBP', 'EUR', 'AUD', 'JPY');
		if (in_array($baseCurrency, $supportedCurrencies) === false)
			throw new Error("Base currency $baseCurrency is currently not supported. Supported values: " . implode("|", $supportedCurrencies));
	}
}

class FiatRateUpdater extends SlpRateUpdater {
	const TOKEN_DECIMALS = 8;
	
	public function getTokenRate(string $tickerUrl, string $baseCurrency): float {
		$this->checkSupportedCurrency($baseCurrency);
		// we could also fetch their USDH ticker https://api.cryptophyl.com/products/BCH-USDH/ticker
		$res = $this->httpGet($tickerUrl);
		if (empty($res)) {
			return -1.0;
		}
		$json = json_decode($res);
		if ($json === null || empty($json->price)) {
			return -2.0;
		}
		if ($json->price > 0.0) {
			$bchRate = $this->getBitcoinCashRate($baseCurrency);
			if ($bchRate <= 0.0)
				return -3.0;
			$shift = pow(10, static::TOKEN_DECIMALS);
			return floor($bchRate * $json->price * $shift) / $shift;
		}
		return -4.0;
	}
}
?>